/*
 * ArpTelnet Copyright (c)1997-98 by Dianne Hackborn.
 * All rights reserved.
 *
 * Based (increasingly loosly) on WebTerm, a Web-based terminal
 * applet written in Java, which is
 * Copyright (C)1996 by the National Alliance for Computational
 * Science and Engineering (NACSE).
 * See <URL:http://www.nacse.org/> for more information.
 *
 * This code is not public domain, nor freely distributable.
 * Please direct any questions or requests to Dianne Hackborn,
 * at <hackbod@lucent.com> or <hackbod@angryredplanet.com>.
 *
 */

//#define ArpDEBUG 1

#ifndef _APPLICATION_H
#include <be/app/Application.h>
#endif

#ifndef _SCREEN_H
#include <be/interface/Screen.h>
#endif

#ifndef _RECT_H
#include <be/interface/Rect.h>
#endif

#ifndef _ALERT_H
#include <be/interface/Alert.h>
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

#ifndef ARPKERNEL_ARPMULTIDIR_H
#include <ArpKernel/ArpMultiDir.h>
#endif

#ifndef ARPNETWORK_ARPTELNET_H
#include <ArpNetwork/ArpTelnet.h>
#endif

#ifndef ARPKERNEL_ARPSTRING_H
#include <ArpKernel/ArpString.h>
#endif

#ifndef MAIN_H
#include "Main.h"
#endif

#ifndef TERMINALWIN_H
#include "TerminalWin.h"
#endif

#include <stdlib.h>

#if ArpDEBUG
#include <ArpCollections/ArpSTLVector.h>

static void ShowURLParts(const ArpString& url)
{
	ArpString proto, addr, path, loc;
	ArpSTLVector<ArpString> params;
	
	DB(DBALL, cdb << "--- Parsing URL: " << url << endl);
	status_t err = ArpParseURL(url, &proto, &addr, &path, &params, &loc);
	if( err ) {
		DB(DBALL, cdb << "Error during parse!  Code: " << err << endl);
	}
	if( !proto.IsNull() ) {
		DB(DBALL, cdb << "Protocol: " << proto << endl);
	} else {
		 DB(DBALL, cdb << "No protocol specified." << endl);
	}
	if( !addr.IsNull() ) {
		DB(DBALL, cdb << "Address: " << addr << endl);
	} else {
		DB(DBALL, cdb << "No address specified." << endl);
	}
	if( !path.IsNull() ) {
		DB(DBALL, cdb << "Path: " << path << endl);
	} else {
		DB(DBALL, cdb << "No path specified." << endl);
	}
	
	if( params.size() > 0 ) {
		for( int32 i = 0; i<params.size(); i++ ) {
			DB(DBALL, cdb << "Parameter #" << i << ": " << params[i] << endl);
		}
	} else {
		DB(DBALL, cdb << "No parameters specified." << endl);
	}
	
	if( !loc.IsNull() ) {
		DB(DBALL, cdb << "Location: " << loc << endl);
	} else {
		DB(DBALL, cdb << "No location specified." << endl);
	}
}
#endif

main()
{	
	MainApp *myApplication;

	myApplication = new MainApp();

#if ArpDEBUG
	// Here is an example using ArpMultiDir...  originally written
	// just for my own testing, but I thought I'd leave it in since
	// the class is otherwise rather undocumented. ;)
	BPath path;
	if( find_directory(B_BEOS_ADDONS_DIRECTORY,&path) == B_OK ) {
		printf("BEOS ADDONS: %s\n",path.Path());
	}
	if( find_directory(B_COMMON_ADDONS_DIRECTORY,&path) == B_OK ) {
		printf("COMMON ADDONS: %s\n",path.Path());
	}
	if( find_directory(B_USER_ADDONS_DIRECTORY,&path) == B_OK ) {
		printf("USER ADDONS: %s\n",path.Path());
	}
	
	ArpMultiDir mdir;
	mdir.AddDirectory(B_BEOS_ADDONS_DIRECTORY,"ARP/TermEmulator");
	mdir.AddDirectory(B_COMMON_ADDONS_DIRECTORY,"ARP/TermEmulator");
	mdir.AddDirectory(B_USER_ADDONS_DIRECTORY,"ARP/TermEmulator");
	mdir.AddEnvVar("ADDON_PATH","ARP/TermEmulator");
	BEntry entry;
	mdir.Rewind();
	while( mdir.GetNextEntry(&entry,true) == B_NO_ERROR ) {
		BPath path;
		entry.GetPath(&path);
		printf("In dir %s found %s\n","<unknown>",path.Path());
	}
	
	ShowURLParts("telnet:");
	ShowURLParts("telnet://adam.enteract.com/");
	ShowURLParts("telnet://adam.enteract.com:23/");
	ShowURLParts("telnet://adam.enteract.com:23/%3Bhackbod");
	ShowURLParts("www.enteract.com");
	ShowURLParts("http://www.enteract.com");
	ShowURLParts("http://www.enteract.com/");
	ShowURLParts("http:/root/dir");
	ShowURLParts("http:relative/dir");
	ShowURLParts("http:relative/dir?with=param?more=param2");
	ShowURLParts("/root/dir");
	ShowURLParts("relative/dir");
	ShowURLParts("relative/dir?with=param?more=param2");
	ShowURLParts("http://www.enteract.com/full/url?with=param?more=param2");
	ShowURLParts("#piece");
	ShowURLParts("http://www.enteract.com/full/url?with=param?more=param2#piece");
#endif

	myApplication->Run();
	
	delete(myApplication);
	return(0);
}

MainApp::MainApp()
	: BApplication("application/x-vnd.ARP-ArpTelnet"),
	  cmdline(false), mMadeWindow(false)
{
}

bool MainApp::MakeTermWindow(const char* url)
{
	TerminalWin* termwin;
	DB(DBALL, cerr << "Creating terminal window..." << endl);
	termwin = new TerminalWin(BPoint(20,20),
								"ArpTelnet v" PROGRAM_VERSION,
								"About ArpTelnet" B_UTF8_ELLIPSIS);
	
	// make window visible
	if(termwin) {
		DB(DBALL, cerr << "Sizing terminal window..." << endl);
		termmsg = BMessenger(termwin);
		BScreen screen(termwin);
		if( screen.IsValid() ) {
			BRect wfrm = termwin->Frame();
			BRect sfrm = screen.Frame();
			float height = wfrm.Height();
			float width = wfrm.Width();
			if( height > sfrm.Height() ) height = sfrm.Height();
			if( width > sfrm.Width() ) width = sfrm.Width();
			termwin->MoveTo( sfrm.left + (sfrm.Width()-width)/2,
								sfrm.top + (sfrm.Height()-height)/2 );
			termwin->ResizeTo(width,height);
		}
		DB(DBALL, cerr << "Creating remote to: " << url << endl);
		if( url ) termwin->SetRemote(url);
		DB(DBALL, cerr << "Showing terminal window..." << endl);
		termwin->Show();
		return true;
	}
	
	return false;
}

void MainApp::AboutRequested(void)
{
	BAlert* alert =
		new BAlert("About ArpTelnet",
					"ArpTelnet version " PROGRAM_VERSION
					" / " __DATE__ "\n"
					B_UTF8_COPYRIGHT "1997-98 Dianne Hackborn.\n\n"
					"A terminal class featuring add-on emulators and "
					"a TELNET client.  More "
					"information is available from Dianne Hackborn "
					"<hackbod@angryredplanet.com>.\n\n"
					"This program and associated code is NOT public domain.  "
					"You may freely use the application however you would "
					"like, and freely use the associated program classes in "
					"your own programs.\n\n"
					"NOTE: This is a PRE-RELEASE of the next version of "
					"ArpTelnet.  I am release this early simply so that there "
					"is a version available under R4.  From a bit of use, it "
					"seems to be working okay for me, but no promises"
					B_UTF8_ELLIPSIS,
					"Fantastic");
	if( alert ) {
		alert->Go(NULL);
	}
}

void MainApp::DispatchMessage(BMessage *message,
								  BHandler *handler)
{
	DB(DBALL, cerr << "MainApp::DispatchMessage: " <<
				*message << endl);
	inherited::DispatchMessage(message,handler);
}

void MainApp::ArgvReceived(int32 argc, char** argv)
{
	// Process command line arguments:  Connect to given host
	// and optional port.
	if( argc >= 2 ) {
		BMessage connmsg(ARPTELNET_CONNECT_MSG);
		if( strchr(argv[1],':') != NULL ) {
			if( !mMadeWindow ) {
				mMadeWindow = MakeTermWindow(argv[1]);
			} else {
				// Send a raw url.
				connmsg.AddString("url", argv[1]);
				termmsg.SendMessage(&connmsg);
			}
		} else {
			ArpString url;
			url = ArpLIT("telnet://") + argv[1];
			if( argc >= 3 ) url += ArpLIT(":") + argv[2] + ":";
			if( !mMadeWindow ) {
				mMadeWindow = MakeTermWindow(url);
			} else {
				connmsg.AddString("url", url);
				termmsg.SendMessage(&connmsg);
			}
		}
		cmdline = true;
	}
}

void MainApp::ReadyToRun()
{
	if( !mMadeWindow ) mMadeWindow = MakeTermWindow("shell:");
}

